// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include "graid.h"
#include "dmapool.h"

#include <linux/mutex.h>
#include <linux/completion.h>

#define DEVICE_NAME "graid"
#define CLASS_NAME "graid"

struct device* graid_device = NULL;
int marked_clean_shutdown = 0;
u64 graid_dma_mask;

static int major;
static struct class*  graid_class  = NULL;

static int graid_create_chrdev(void)
{
	int rc;

	rc = register_chrdev(0, DEVICE_NAME, &graid_fops);
	if (rc < 0) {
		printk(KERN_ALERT "Registering char device failed with %d\n", rc);
		return rc;
	}
	major = rc;

	graid_class = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(graid_class)) {
		printk(KERN_ALERT "Failed to register device class\n");
		rc = PTR_ERR(graid_class);
		goto unreg_out;
	}

	graid_device = device_create(graid_class, NULL, MKDEV(major, 0), NULL, DEVICE_NAME);
	if (IS_ERR(graid_device)) {
		printk(KERN_ALERT "Failed to create the device\n");
		rc = PTR_ERR(graid_device);
		goto class_des_out;
	}
	graid_device->dma_mask = &graid_dma_mask;

	rc = dma_set_mask_and_coherent(graid_device, 0xFFFFFFFFFFFFu);
	if (rc) {
		printk(KERN_ALERT "Failed set DMA mask\n");
		goto device_des_out;
	}

	//Assume that DMA address range are checked to be correct
	rc = init_dma_pools(graid_device);
	if (rc) {
		printk(KERN_ALERT "Failed to initialize DMA pools\n");
		goto device_des_out;
	}
        return 0;

device_des_out:
	device_destroy(graid_class, MKDEV(major, 0));
class_des_out:
	class_destroy(graid_class);
unreg_out:
	unregister_chrdev(major, DEVICE_NAME);
	return rc;
}

static void graid_destroy_chrdev(void)
{
	cleanup_dma_pools(graid_device);
	device_destroy(graid_class, MKDEV(major, 0));
	class_destroy(graid_class);
	unregister_chrdev(major, DEVICE_NAME);
}

static int __init graid_init(void)
{
	int rc;

	rc = graid_nvidia_init();
	if (rc)
		return rc;

	rc = graid_init_dg_queue();
	if (rc)
		goto err_exit_nvidia;

	rc = graid_create_chrdev();
	if (rc)
		goto err_clean_dg_queue;

	rc = graid_pci_nvme_init();
	if (rc)
		goto err_destroy_chrdev;

	rc = graid_vir_nvme_init();
	if (rc)
		goto err_exit_pci;

	rc = graid_blk_init();
	if (rc)
		goto err_exit_vir;

	pr_info("GRAID module loaded\n");
	return 0;

err_exit_vir:
	graid_vir_nvme_exit();
err_exit_pci:
	graid_pci_nvme_exit();
err_destroy_chrdev:
	graid_destroy_chrdev();
err_clean_dg_queue:
	graid_cleanup_dg_queue();
err_exit_nvidia:
	graid_nvidia_exit();

	return rc;
}

static void __exit graid_exit(void)
{
	graid_blk_exit();
	graid_vir_nvme_exit();
	graid_pci_nvme_exit();
	graid_destroy_chrdev();
	graid_cleanup_dg_queue();
	graid_nvidia_exit();
	pr_info("GRAID module unloaded\n");
}

MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
module_init(graid_init);
module_exit(graid_exit);
