// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include "graid.h"
#include "ioctl.h"

#include <linux/delay.h>

static bool opened = false;
static DEFINE_SPINLOCK(open_lock);
static DEFINE_MUTEX(ioctl_mutex);
static int ioctl_reader_cnt = 0;
static int ioctl_writer_active = 0;

static int graid_open(struct inode *inode, struct file *filp)
{
	unsigned long flags;

	spin_lock_irqsave(&open_lock, flags);
	if (opened) {
		spin_unlock_irqrestore(&open_lock, flags);
		pr_debug("Graid already opened\n");
		return -EBUSY;
	}
	opened = true;
	spin_unlock_irqrestore(&open_lock, flags);

	if (!graid_vir_nvme_ready()) {
		pr_warn("Graid not ready; some VD not released yet\n");
		opened = false;
		return -EAGAIN;
	}

	if (!try_module_get(THIS_MODULE)) {
		pr_err("Failed to get module reference\n");
		opened = false;
		return -ENODEV;
	}

	pr_debug("Graid opened successfully\n");
	return 0;
}

static int graid_release(struct inode *inode, struct file *filp)
{
	graid_vir_nvme_cleanup();
	graid_blk_fwd_cleanup();
	graid_unpin_and_release();
	module_put(THIS_MODULE);
	opened = false;

	pr_debug("graid released successfully\n");
	return 0;
}

static inline
long graid_ioctl_get_nr_cpus(void __user* nr_cpus)
{
	unsigned long cpunr = graid_online_cores();
	if (copy_to_user(nr_cpus, &cpunr, sizeof(cpunr)))
		return -EFAULT;
	return 0;
}

void ioctl_read_lock(void)
{
	mutex_lock(&ioctl_mutex);
	while (ioctl_writer_active) {
		mutex_unlock(&ioctl_mutex);
		msleep(1);
		mutex_lock(&ioctl_mutex);
	}
	++ioctl_reader_cnt;
	mutex_unlock(&ioctl_mutex);
}

void ioctl_read_unlock(void)
{
	mutex_lock(&ioctl_mutex);
	--ioctl_reader_cnt;
	mutex_unlock(&ioctl_mutex);
}

void ioctl_write_lock(void)
{
	mutex_lock(&ioctl_mutex);
	while (ioctl_reader_cnt || ioctl_writer_active) {
		mutex_unlock(&ioctl_mutex);
		msleep(1);
		mutex_lock(&ioctl_mutex);
	}
	ioctl_writer_active = 1;
	mutex_unlock(&ioctl_mutex);
}

void ioctl_write_unlock(void)
{
	mutex_lock(&ioctl_mutex);
	ioctl_writer_active = 0;
	mutex_unlock(&ioctl_mutex);
}

static void graid_do_lock(unsigned int cmd)
{
	switch (cmd) {
	case GRAID_GET_DMA_POOL:
	case GRAID_GET_NR_CPUS:
	case GRAID_DMA_GET_PRE_ALLOC:
	case GRAID_PCI_RESET_NVME:
	case GRAID_PCI_DEALLOC_NVME:
		ioctl_read_lock();
		break;
	default:
		ioctl_write_lock();
	};
}

static void graid_do_unlock(unsigned int cmd)
{
	switch (cmd) {
	case GRAID_GET_DMA_POOL:
	case GRAID_GET_NR_CPUS:
	case GRAID_DMA_GET_PRE_ALLOC:
	case GRAID_PCI_RESET_NVME:
	case GRAID_PCI_DEALLOC_NVME:
		ioctl_read_unlock();
		break;
	default:
		ioctl_write_unlock();
	};
}

static long graid_ioctl(struct file* file, unsigned int cmd, unsigned long arg)
{
	int rc = -EINVAL;;
	graid_do_lock(cmd);
	switch (cmd) {
	case GRAID_GET_DMA_POOL:
		rc = graid_ioctl_get_dma_pool((void __user*)arg);
		break;
	case GRAID_GET_NR_CPUS:
		rc = graid_ioctl_get_nr_cpus((void __user*)arg);
		break;
	case GRAID_DMA_GET_PRE_ALLOC:
		rc = graid_ioctl_get_mem_addrs((void __user*)arg);
		break;
	case GRAID_DMA_SET_BAR1:
		rc = graid_ioctl_dma_set_bar1((void __user*)arg);
		break;
	case GRAID_DMA_RESET_BAR1:
		rc = graid_ioctl_dma_reset_bar1((void __user*)arg);
		break;
	case GRAID_PCI_QUERY_NVME:
		rc = graid_pci_nvme_query((struct graid_pci_nvme_dev __user *)arg);
		break;
	case GRAID_PCI_RESET_NVME:
		rc = graid_pci_nvme_reset(arg);
		break;
	case GRAID_PCI_ALLOC_ID:
		rc = graid_pci_alloc_id((void __user*)arg);
		break;
	case GRAID_PCI_FREE_ID:
		rc = graid_pci_free_id(arg);
		break;
	case GRAID_PCI_DEALLOC_NVME:
		rc = graid_pci_nvme_dealloc(arg);
		break;
	case GRAID_DG_CREATE:
		rc = graid_create_dg((struct graid_nvme_drive_group __user *)arg);
		break;
	case GRAID_DG_DESTROY:
		rc = graid_destroy_dg(arg);
		break;
	case GRAID_VIR_NVME_CREATE:
		rc = graid_create_ctrl((struct graid_virt_nvme_dev __user *)arg);
		break;
	case GRAID_VIR_NVME_DESTROY:
		rc = graid_destroy_ctrl(arg);
		break;
	case GRAID_VIR_NVME_RESUME:
		{
			u32 dgid = arg & 0xFFFFFFFFllu;
			u32 optimal = arg >> 32;
			rc = graid_start_queues(dgid, optimal);
		}
		break;
	case GRAID_VIR_NVME_PAUSE:
		rc = graid_stop_queues(arg);
		break;
	case GRAID_VIR_NVME_CHECK:
		rc = graid_check_ctrl(arg);
		break;
	case GRAID_BLK_CREATE:
		rc = graid_blk_create((void __user *)arg);
		break;
	case GRAID_BLK_DESTROY:
		rc = graid_blk_destroy(arg);
		break;
	case GRAID_BLK_RESET:
		rc = graid_blk_reset(arg);
		break;
	case GRAID_GPU_PIN_BAR1:
		rc = graid_gpu_pin_bar1((struct graid_gpu_bar1_req*)arg);
		break;
	case GRAID_GPU_UNPIN_BAR1:
		rc = graid_gpu_unpin_bar1((struct graid_gpu_bar1_req*)arg);
		break;
	case GRAID_CLEAN_SHUTDOWN:
		marked_clean_shutdown = 1;
		break;
	default:
		break;
	};
	graid_do_unlock(cmd);
	return rc;
}

void graid_vma_open(struct vm_area_struct *vma)
{
	pr_debug("graid VMA open, virt %lx, phys %lx\n",
            vma->vm_start, vma->vm_pgoff << PAGE_SHIFT);
}

void graid_vma_close(struct vm_area_struct *vma)
{
	pr_debug("graid VMA close, virt %lx, phys %lx\n",
            vma->vm_start, vma->vm_pgoff << PAGE_SHIFT);
}

static const struct vm_operations_struct graid_mmap_ops = {
	.open = graid_vma_open,
	.close = graid_vma_close,
};

static int graid_valid_mapping_address(uint64_t addr, uint64_t len)
{
	int rc;
	//For p2pdma_check
	//if (addr == 0xf2710000 && len == 0x4000)
	//	return 1;
	rc = graid_is_dma_pool(addr, len);
	if (rc)
		return rc;
	if ((addr & 0x3FFFFFFFllu) == 0 && len == 0x40000000llu)
		return 1;
	return graid_is_pci_nvme_bar0(addr, len) ? 2 : 0;
}

static int graid_mmap(struct file *file, struct vm_area_struct *vma)
{
	size_t size = vma->vm_end - vma->vm_start;
	size_t start = vma->vm_pgoff << PAGE_SHIFT;
	int rc;

	//printk(KERN_INFO "%s: %lx -> %lx\n", __func__,
	//		start, start + size);
	rc = graid_valid_mapping_address(start, size);
	if (!rc) {
		pr_err("Unsupported mapping range.\n");
		return -EINVAL;
	}

	if (rc >= GRAID_DMA_NODE_OFFSET) {
		rc = graid_map_dmapool_pages(vma, rc - GRAID_DMA_NODE_OFFSET, start);
	} else {
		if (rc == 2)
			vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
		rc = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
				     size, vma->vm_page_prot);
		vma->vm_ops = &graid_mmap_ops;
		graid_vma_open(vma);
	}
	return rc;
}

struct file_operations graid_fops = {
	.open           = graid_open,
	.release        = graid_release,
	.unlocked_ioctl = graid_ioctl,
	.mmap           = graid_mmap
};
